import json

from django.core.exceptions import ValidationError
from django.db.models import Count, Sum, ExpressionWrapper, FloatField
from django.db.models.functions import Coalesce, NullIf
from django.forms import ChoiceField
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django_datatables.datatables import DatatableTable
from django_menus.menu import MenuItem
from django_modals.fields import FieldEx
from django_modals.helper import show_modal
from django_modals.modals import Modal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.colour_picker import ColourPickerWidget
from django_modals.widgets.select2 import Select2Multiple
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.columns import ReportBuilderNumberColumn
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.globals import NUMBER_FIELDS, ANNOTATION_CHOICE_SUM, \
    ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT
from advanced_report_builder.models import SingleValueReport, ReportType
from advanced_report_builder.utils import get_field_details
from advanced_report_builder.variable_date import VariableDate
from advanced_report_builder.views.charts_base import ChartBaseView
from advanced_report_builder.views.datatables.modal import TableFieldModal, TableFieldForm
from advanced_report_builder.views.datatables.utils import TableUtilsMixin
from advanced_report_builder.views.modals_base import QueryBuilderModalBase


class SingleValueView(ChartBaseView):
    number_field = ReportBuilderNumberColumn
    template_name = 'advanced_report_builder/single_values/report.html'
    use_annotations = False

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.singlevaluereport
        return super().dispatch(request, *args, **kwargs)

    def _process_aggregations(self, fields, aggregations_type=ANNOTATION_CHOICE_SUM, divider=None):
        field = self.chart_report.field
        base_model = self.chart_report.get_base_modal()
        report_builder_class = getattr(base_model, self.chart_report.report_type.report_builder_class_name, None)

        django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                  field=field,
                                                                  report_builder_class=report_builder_class)

        if (isinstance(django_field, NUMBER_FIELDS) or
                col_type_override is not None and col_type_override.annotations):

            self.get_number_field(annotations_type=aggregations_type,
                                  index=0,
                                  data_attr={},
                                  table_field={'field': field, 'title': field},
                                  fields=fields,
                                  col_type_override=col_type_override,
                                  decimal_places=self.chart_report.decimal_places,
                                  convert_currency_fields=True,
                                  divider=divider)
        else:
            raise ReportError('not a number field')

    def _get_count(self, fields):

        number_function_kwargs = {'aggregations': {'count': Count(1)},
                                  'field': 'count',
                                  'column_name': 'count',
                                  'options': {'calculated': True},
                                  'model_path': ''}

        field = self.number_field(**number_function_kwargs)
        fields.append(field)

    def get_percentage_field(self, fields,
                             numerator_field_name, numerator_col_type_override,
                             denominator_field_name, denominator_col_type_override,
                             numerator_filter, decimal_places=0):
        if numerator_col_type_override:
            actual_numerator_field_name = numerator_col_type_override.field
        else:
            actual_numerator_field_name = numerator_field_name

        if denominator_col_type_override:
            actual_denominator_field_name = denominator_col_type_override.field
        else:
            actual_denominator_field_name = denominator_field_name

        number_function_kwargs = {}
        if decimal_places:
            number_function_kwargs = {'decimal_places': int(decimal_places)}

        new_field_name = f'{actual_numerator_field_name}_{actual_denominator_field_name}'

        if numerator_col_type_override is not None and numerator_col_type_override.annotations:

            if not isinstance(numerator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(numerator_col_type_override.annotations.values())[0]
            if numerator_filter and isinstance(annotations, (Sum, Count)):
                annotations.filter = numerator_filter
                numerator = Coalesce(annotations + 0.0, 0.0)
            else:
                numerator = Coalesce(annotations + 0.0, 0.0)
        else:
            if numerator_filter:
                numerator = Coalesce((Sum(actual_numerator_field_name, filter=numerator_filter)+0.0), 0.0)
            else:
                numerator = Coalesce((Sum(actual_numerator_field_name) + 0.0), 0.0)

        if denominator_col_type_override is not None and denominator_col_type_override.annotations:
            if not isinstance(denominator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(denominator_col_type_override.annotations.values())[0]

            denominator = Coalesce(annotations + 0.0, 0.0)
        else:
            denominator = Coalesce(Sum(actual_denominator_field_name) + 0.0, 0.0)

        number_function_kwargs['aggregations'] = {
            new_field_name: ExpressionWrapper((numerator / NullIf(denominator, 0)) * 100.0, output_field=FloatField())}
        field_name = new_field_name

        number_function_kwargs.update({'field': field_name,
                                       'column_name': field_name,
                                       'options': {'calculated': True},
                                       'model_path': ''})

        field = self.number_field(**number_function_kwargs)

        fields.append(field)

        return field_name

    def _process_percentage(self, fields):
        denominator_field = self.chart_report.field
        numerator_field = self.chart_report.numerator
        base_model = self.chart_report.get_base_modal()
        report_builder_class = getattr(base_model, self.chart_report.report_type.report_builder_class_name, None)

        deno_django_field, denominator_col_type_override, _, _ = get_field_details(
            base_model=base_model,
            field=denominator_field,
            report_builder_class=report_builder_class)
        if (not isinstance(deno_django_field, NUMBER_FIELDS) and
                (denominator_col_type_override is not None and not denominator_col_type_override.annotations)):
            raise ReportError('denominator is not a number field')

        num_django_field, numerator_col_type_override, _, _ = get_field_details(
            base_model=base_model,
            field=denominator_field,
            report_builder_class=report_builder_class)

        if (not isinstance(num_django_field, NUMBER_FIELDS) and
                (numerator_col_type_override is not None and not numerator_col_type_override.annotations)):
            raise ReportError('numerator is not a number field')

        numerator_filter = None
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            numerator_filter = self.process_filters(search_filter_data=report_query.extra_query)

        self.get_percentage_field(fields=fields,
                                  numerator_field_name=numerator_field,
                                  numerator_col_type_override=numerator_col_type_override,
                                  denominator_field_name=denominator_field,
                                  denominator_col_type_override=denominator_col_type_override,
                                  numerator_filter=numerator_filter,
                                  decimal_places=self.chart_report.decimal_places)

    def _process_percentage_from_count(self, fields):
        report_query = self.get_report_query(report=self.chart_report)

        numerator_filter = None
        if report_query:
            numerator_filter = self.process_filters(search_filter_data=report_query.extra_query)

        if numerator_filter:
            numerator = Coalesce(Count(1, filter=numerator_filter) + 0.0, 0.0)
        else:
            numerator = Coalesce(Count(1) + 0.0, 0.0)
        denominator = Coalesce(Count(1) + 0.0, 0.0)
        a = (numerator / denominator) * 100.0

        number_function_kwargs = {'aggregations': {'count': a},
                                  'field': 'count',
                                  'column_name': 'count',
                                  'options': {'calculated': True},
                                  'model_path': ''}
        decimal_places = self.chart_report.decimal_places
        if decimal_places:
            number_function_kwargs['decimal_places'] = int(decimal_places)

        field = self.number_field(**number_function_kwargs)
        fields.append(field)

    def process_query_results(self, base_model, table):
        single_value_type = self.chart_report.single_value_type
        fields = []
        if single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_COUNT:
            self._get_count(fields=fields)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_SUM:
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_COUNT_AND_SUM:
            self._get_count(fields=fields)
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_FROM_COUNT:
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME:
            divider = self.get_period_divider(annotation_value_choice=self.chart_report.average_scale,
                                              start_date_type=self.chart_report.average_start_period,
                                              end_date_type=self.chart_report.average_end_period)
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM, divider=divider)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_PERCENT:
            self._process_percentage(fields=fields)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT:
            self._process_percentage_from_count(fields=fields)
        return fields

    def set_prefix(self):
        self.table.prefix = self.chart_report.prefix
        if self.table.prefix:
            self.table.prefix += ' '

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.set_prefix()
        self.table.single_value = self.chart_report
        self.table.enable_links = self.kwargs.get('enable_links')
        self.table.datatable_template = 'advanced_report_builder/single_values/middle.html'
        self.table.breakdown_url = self.get_breakdown_url()
        context['single_value_report'] = self.chart_report
        return context

    def get_breakdown_url(self):
        if self.table.single_value.show_breakdown:
            return show_modal('advanced_report_builder:show_breakdown_modal',
                              'pk-',
                              self.table.single_value.id,
                              '-enable_links-',
                              self.table.enable_links, href=True)
        return None

    def pod_dashboard_view_menu(self):
        return []

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [MenuItem(f'advanced_report_builder:single_value_modal,pk-{chart_report_id}{slug_str}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary']),
                *self.duplicate_menu(request=self.request, report_id=chart_report_id)]

    def queries_menu(self):
        return []


class SingleValueModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/single_values/modal.html'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = SingleValueReport
    widgets = {'tile_colour': ColourPickerWidget,
               'report_tags': Select2Multiple,
               'show_breakdown': Toggle(attrs={'data-onstyle': 'success', 'data-on': 'YES', 'data-off': 'NO'})}

    form_fields = ['name',
                   'notes',
                   'report_type',
                   'report_tags',
                   ('single_value_type', {'label': 'Value type'}),
                   ('numerator', {'label': 'Numerator field'}),
                   'average_scale',
                   'average_start_period',
                   'average_end_period',
                   'field',
                   'prefix',
                   'tile_colour',
                   ('decimal_places', {'field_class': 'col-md-5 col-lg-3 input-group-sm'}),
                   'show_breakdown',
                   'breakdown_fields'
                   ]

    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger('single_value_type', 'onchange', [
            {'selector': '#div_id_field',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_COUNT: 'hide',
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT: 'hide'},
             'default': 'show'},
            {'selector': '#div_id_prefix',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_COUNT: 'hide',
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT: 'hide'},
             'default': 'show'},
            {'selector': '#div_id_numerator',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_PERCENT: 'show'},
             'default': 'hide'},
            {'selector': '#div_id_extra_query_data',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_PERCENT: 'show',
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT: 'show'},
             'default': 'hide'},
            {'selector': 'label[for=id_field]',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_PERCENT: ('html', 'Denominator field')},
             'default': ('html', 'Field')},

            {'selector': '#div_id_average_scale',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
             'default': 'hide'},
            {'selector': '#div_id_average_start_period',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
             'default': 'hide'},
            {'selector': '#div_id_average_end_period',
             'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
             'default': 'hide'},
        ])

        form.add_trigger('show_breakdown', 'onchange', [
            {'selector': '#div_id_breakdown_fields', 'values': {'checked': 'show'}, 'default': 'hide'},
        ])

        if 'data' in _kwargs:
            field = _kwargs['data'].get('field')
            numerator = _kwargs['data'].get('numerator')
            report_type_id = _kwargs['data'].get('report_type')
            report_type = get_object_or_404(ReportType, id=report_type_id)
        else:
            field = form.instance.field
            report_type = form.instance.report_type
            numerator = form.instance.numerator

        self.setup_field(field_type='number',
                         form=form,
                         field_name='field',
                         selected_field_id=field,
                         report_type=report_type)

        self.setup_field(field_type='number',
                         form=form,
                         field_name='numerator',
                         selected_field_id=numerator,
                         report_type=report_type)

        self.add_query_data(form, include_extra_query=True)
        form.fields['notes'].widget.attrs['rows'] = 3
        url = reverse('advanced_report_builder:single_value_field_modal',
                      kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        range_type_choices = VariableDate.RANGE_TYPE_CHOICES
        form.fields['average_start_period'] = ChoiceField(required=False, choices=range_type_choices)
        form.fields['average_end_period'] = ChoiceField(required=False, choices=range_type_choices)
        return ('name',
                'notes',
                'report_type',
                'report_tags',
                'single_value_type',
                'numerator',
                'field',
                'average_scale',
                'average_start_period',
                'average_end_period',
                'prefix',
                FieldEx('extra_query_data',
                        template='advanced_report_builder/query_builder.html',
                        ),
                'tile_colour',
                'decimal_places',
                'show_breakdown',
                FieldEx('breakdown_fields',
                        template='advanced_report_builder/select_column.html',
                        extra_context={'select_column_url': url}),
                FieldEx('query_data',
                        template='advanced_report_builder/query_builder.html')
                )

    def select2_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='number',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_numerator(self, **kwargs):
        return self.get_fields_for_select2(field_type='number',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    # noinspection PyUnusedLocal
    def clean(self, form, cleaned_data):
        single_value_type = cleaned_data['single_value_type']
        if (single_value_type not in [SingleValueReport.SINGLE_VALUE_TYPE_COUNT,
                                      SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT] and
                not cleaned_data['field']):
            raise ValidationError("Please select a field")

        if single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME:
            if cleaned_data['average_scale'] is None:
                form.add_error('average_scale', "Please select a time scale")
            if cleaned_data['average_start_period'] is None:
                form.add_error('average_start_period', "Please select a start period")
            if cleaned_data['average_end_period'] is None:
                form.add_error('average_end_period', "Please select a end period")

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_fields, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        self._get_fields(base_model=base_model,
                         fields=fields,
                         tables=tables,
                         report_builder_class=report_builder_fields)
        return self.command_response('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))


class ShowBreakdownModal(TableUtilsMixin, Modal):
    button_container_class = 'text-center'
    size = 'xl'

    def modal_title(self):
        return self.table_report.name

    def add_table(self, base_model):
        return DatatableTable(view=self, model=base_model)

    def setup_table(self):
        single_value_report = get_object_or_404(SingleValueReport, pk=self.slug['pk'])
        self.kwargs['enable_links'] = self.slug['enable_links'] == 'True'
        self.table_report = single_value_report
        base_model = single_value_report.get_base_modal()
        table = self.add_table(base_model=base_model)
        table.extra_filters = self.extra_filters
        table_fields = single_value_report.breakdown_fields
        report_builder_class = getattr(base_model,
                                       self.table_report.report_type.report_builder_class_name, None)
        fields_used = set()
        self.process_query_results(report_builder_class=report_builder_class,
                                   table=table,
                                   base_model=base_model,
                                   fields_used=fields_used,
                                   table_fields=table_fields)

        table.ajax_data = False
        table.table_options['pageLength'] = 25
        table.table_options['bStateSave'] = False
        return table

    def modal_content(self):
        table = self.setup_table()
        return table.render()


class SingleValueTableFieldForm(TableFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'save_query_builder_id_query_data'},
                    {'function': 'save_query_builder_id_extra_query_data'},
                    {'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)


class SingleValueTableFieldModal(TableFieldModal):
    form_class = SingleValueTableFieldForm

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command({'function': 'set_attr',
                          'selector': f'#{selector}',
                          'attr': 'data-attr',
                          'val': _attr})

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})
        self.add_command({'function': 'save_query_builder_id_query_data'})
        self.add_command({'function': 'save_query_builder_id_extra_query_data'})
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')



import re

from django.apps import apps

from advanced_report_builder.utils import get_field_details


def get_menu_fields(base_model, report_builder_class,
                    menus=None, codes=None, code_prefix='', next_code_prefix='__',
                    previous_base_model=None, table=None):

    for report_builder_field in report_builder_class.fields:
        django_field, col_type_override, columns, _ = get_field_details(base_model=base_model,
                                                                        field=report_builder_field,
                                                                        report_builder_class=report_builder_class,
                                                                        table=table)
        for column in columns:
            full_id = code_prefix + column.column_name
            if menus is not None:
                menus.append({'code': full_id, 'text': column.title})
            if codes is not None:
                codes.add(full_id)

    for include_field, include in report_builder_class.includes.items():
        app_label, model, report_builder_fields_str = include['model'].split('.')

        new_model = apps.get_model(app_label, model)
        if new_model != previous_base_model:
            new_report_builder_class = getattr(new_model, report_builder_fields_str, None)
            title = new_report_builder_class.title
            if menus is not None:
                menu = []
            else:
                menu = None
            get_menu_fields(base_model=new_model,
                            report_builder_class=new_report_builder_class,
                            menus=menu,
                            codes=codes,
                            code_prefix=f"{code_prefix}{include_field}{next_code_prefix}",
                            next_code_prefix=next_code_prefix,
                            previous_base_model=base_model,
                            table=table)
            if menus is not None:
                menus.append({'text': title, 'menu': menu})


def get_data_merge_variables(html):
    variables = re.findall('{{\s*([^*\s*}}]+)\s*}}', html)

    all_fields = set()
    for variable in variables:
        if '|' in variable:
            field = variable.split('|')[0]
        elif '&' in variable:
            field = variable.split('&')[0]
        else:
            field = variable
        all_fields.add(field)

    variables = re.findall('{%\s*(if|elif|with)\s([^%}]+)\s*%}', html)

    for variable in variables:
        for field in variable[1].split(' '):
            if field != '' and (field == 'as' or field[0] in ['=', '<', '>']):
                break
            if field not in ['', 'not', 'and' 'or'] and field[0] not in ['(', ')', '"', "'"]:
                all_fields.add(field)
    return all_fields


def get_data_merge_columns(base_model, report_builder_class, html, table):
    all_fields = get_data_merge_variables(html)
    columns = set()
    column_map = {}
    for field in all_fields:
        django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                  field=field,
                                                                  report_builder_class=report_builder_class,
                                                                  table=table)
        if django_field is not None or isinstance(col_type_override.field, (list, tuple)):
            if field not in columns and f'.{field}' not in columns:
                columns.add('.' + field)

            field_parts = field.split('__')

            if col_type_override is not None and field_parts[-1] != col_type_override.field:
                column_map[field] = col_type_override.field
    return columns, column_map







import copy

from django.db.models import Q
from django_datatables.plugins.column_totals import ColumnTotals

from advanced_report_builder.columns import ReportBuilderDateColumn
from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, CURRENCY_COLUMNS, LINK_COLUMNS
from advanced_report_builder.globals import DATE_FORMAT_TYPES_DJANGO_FORMAT, ANNOTATION_VALUE_FUNCTIONS
from advanced_report_builder.utils import split_attr, get_field_details, decode_attribute
from advanced_report_builder.views.report_utils_mixin import ReportUtilsMixin


class TableUtilsMixin(ReportUtilsMixin):
    date_field = ReportBuilderDateColumn
    column_totals_class = ColumnTotals

    def __init__(self, *args, **kwargs):
        self.table_report = None
        super().__init__(*args, **kwargs)

    def get_date_field(self, index, col_type_override, table_field, fields):
        data_attr = split_attr(table_field)
        field_name = table_field['field']
        date_format = data_attr.get('date_format')

        if date_format:
            date_format = DATE_FORMAT_TYPES_DJANGO_FORMAT.get(int(date_format))
        title = table_field.get('title')

        annotations_value = int(data_attr.get('annotations_value', 0))
        if col_type_override and annotations_value == 0:
            col_type_override.table = None
            field = copy.deepcopy(col_type_override)

            if title:
                field.title = title
            fields.append(field)
        else:
            if col_type_override:
                field_name = col_type_override.field

            date_function_kwargs = {'title': table_field.get('title'),
                                    'date_format': date_format}

            if annotations_value != 0:
                new_field_name = f'{annotations_value}_{field_name}_{index}'
                function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
                date_function_kwargs['annotations_value'] = {new_field_name: function(field_name)}
                field_name = new_field_name

            date_function_kwargs.update({'field': field_name,
                                         'column_name': field_name})

            field = self.date_field(**date_function_kwargs)
            fields.append(field)

        return field_name

    @staticmethod
    def get_link_field(table_field, fields, col_type_override):
        col_type_override.table = None
        field = copy.deepcopy(col_type_override)
        data_attr = split_attr(table_field)
        link_css = ''
        if 'link_css' in data_attr:
            link_css = decode_attribute(data_attr['link_css'])

        link_html = ''
        if 'link_html' in data_attr:
            link_html = decode_attribute(data_attr['link_html'])

        if link_css or link_html:
            field.setup_link(link_css=link_css, link_html=link_html)

        is_icon = data_attr.get('is_icon')
        if is_icon == '1':
            field.title = ''
            field.options['no_col_search'] = True
            field.column_defs['orderable'] = False
            field.column_defs['width'] = '10px'

        elif 'title' in table_field:
            field.title = table_field['title']
        fields.append(field)

    def process_query_results(self, report_builder_class, table, base_model,
                              fields_used, table_fields, pivot_fields=None):
        first_field_name = None

        field_name = None
        fields = []
        totals = {}

        has_annotations = False
        if not table_fields:
            return fields, totals, first_field_name

        for index, table_field in enumerate(table_fields):
            field = table_field['field']
            field_attr = {}
            if 'title' in table_field:
                field_attr['title'] = table_field['title']

            fields_used.add(field)
            django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                      field=field,
                                                                      table=table,
                                                                      report_builder_class=report_builder_class,
                                                                      field_attr=field_attr)

            if isinstance(django_field, DATE_FIELDS):
                field_name = self.get_date_field(index=index,
                                                 col_type_override=col_type_override,
                                                 table_field=table_field,
                                                 fields=fields)
            elif isinstance(django_field, NUMBER_FIELDS) and (django_field is None or django_field.choices is None):
                data_attr = split_attr(table_field)

                annotations_type = int(data_attr.get('annotations_type', 0))
                if annotations_type != 0:
                    has_annotations = True
                decimal_places = data_attr.get('decimal_places')

                if annotations_type != 0 and data_attr.get('multiple_columns') == '1':
                    query = self.extra_filters(query=table.model.objects)
                    multiple_column_field = data_attr.get('multiple_column_field')

                    field_report_builder_class = self._get_report_builder_class(
                        base_model=base_model,
                        field_str=multiple_column_field,
                        report_builder_class=report_builder_class)
                    _fields = field_report_builder_class.default_multiple_column_fields
                    default_multiple_column_fields = [multiple_column_field + '__' + x for x in _fields]
                    order_by = f'{multiple_column_field}__{field_report_builder_class.default_multiple_pk}'

                    results = query.order_by(order_by).\
                        distinct(multiple_column_field).values(multiple_column_field, *default_multiple_column_fields)

                    for multiple_index, result in enumerate(results):
                        suffix = self._set_multiple_title(database_values=result,
                                                          value_prefix=multiple_column_field,
                                                          fields=_fields,
                                                          text=field_report_builder_class.default_multiple_column_text)
                        extra_filter = Q((multiple_column_field, result[multiple_column_field]))

                        field_name = self.get_number_field(annotations_type=annotations_type,
                                                           index=f'{index}_{multiple_index}',
                                                           data_attr=data_attr,
                                                           table_field=table_field,
                                                           fields=fields,
                                                           totals=totals,
                                                           col_type_override=col_type_override,
                                                           extra_filter=extra_filter,
                                                           title_suffix=suffix,
                                                           decimal_places=decimal_places)
                else:
                    field_name = self.get_number_field(annotations_type=annotations_type,
                                                       index=index,
                                                       data_attr=data_attr,
                                                       table_field=table_field,
                                                       fields=fields,
                                                       totals=totals,
                                                       col_type_override=col_type_override,
                                                       decimal_places=decimal_places)

            elif isinstance(col_type_override, LINK_COLUMNS):
                self.get_link_field(table_field=table_field,
                                    col_type_override=col_type_override,
                                    fields=fields)
            else:
                data_attr = split_attr(table_field)
                if col_type_override is not None:
                    if data_attr.get('annotation_label') == '1':
                        if isinstance(col_type_override.field, list):
                            table.initial_values += col_type_override.field
                        else:
                            table.initial_values.append(col_type_override.field)

                field_name = field
                if isinstance(col_type_override, CURRENCY_COLUMNS) and totals is not None:

                    css_class = col_type_override.column_defs.get('className')
                    show_total = data_attr.get('show_totals')
                    if show_total == '1':
                        totals[field_name] = {'sum': 'to_fixed', 'decimal_places': 2,
                                              'css_class': css_class}
                if field_attr:
                    field = (field, field_attr)
                fields.append(field)
            if not first_field_name:
                first_field_name = field_name

        if not has_annotations and len(report_builder_class.default_columns) > 0:
            table.add_columns(*report_builder_class.default_columns)
        table.add_columns(*fields)
        table.show_pivot_table = False
        if pivot_fields is not None:
            for pivot_field in pivot_fields:
                pivot_field_data = self._get_pivot_details(base_model=base_model,
                                                           pivot_str=pivot_field['field'],
                                                           report_builder_class=report_builder_class)
                if pivot_field_data is None:
                    continue

                if pivot_field_data['id'] not in fields_used:
                    table.add_columns('.' + pivot_field_data['id'])
                    fields_used.add(pivot_field_data['id'])

                pivot_field_details = pivot_field_data['details']
                table.add_js_filters(pivot_field_details['type'],
                                     pivot_field_data['id'],
                                     filter_title=pivot_field_details['title'],
                                     **pivot_field_details['kwargs'])
                table.show_pivot_table = True

        if totals:
            totals[first_field_name] = {'text': 'Totals'}
            table.add_plugin(self.column_totals_class, totals)

    @staticmethod
    def _set_multiple_title(database_values, value_prefix, fields, text):
        results = {}
        for field in fields:
            value = database_values[value_prefix + '__' + field]
            results[field] = value
        return text.format(**results)

    def extra_filters(self, query):
        report_query = self.get_report_query(report=self.table_report)
        if report_query:
            query = self.process_query_filters(query=query,
                                               search_filter_data=report_query.query)
        return query



from django.views.generic import TemplateView
from django_menus.menu import MenuMixin, MenuItem
from django_modals.modals import ModelFormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.select2 import Select2Multiple

from advanced_report_builder.models import CustomReport
from advanced_report_builder.utils import split_slug


class CustomBaseView(MenuMixin, TemplateView):

    def __init__(self, **kwargs):
        self.report = None
        self.slug = None
        self.enable_edit = None
        self.dashboard_report = None
        self.show_toolbar = False
        super().__init__(**kwargs)

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.slug = split_slug(kwargs.get('slug'))
        self.enable_edit = kwargs.get('enable_edit')
        self.dashboard_report = kwargs.get('dashboard_report')
        if self.enable_edit or (self.dashboard_report and not self.dashboard_report.top) or not self.dashboard_report:
            self.show_toolbar = True
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['report'] = self.report
        context['show_toolbar'] = self.show_toolbar
        context['title'] = self.get_title()
        return context

    def setup_menu(self):
        super().setup_menu()
        if not self.show_toolbar:
            return

        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()

        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu()
        )

    def pod_dashboard_edit_menu(self):
        return [MenuItem(f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary'])]

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    # noinspection PyMethodMayBeStatic
    def pod_report_menu(self):
        return []

    def get_title(self):
        if self.dashboard_report and self.dashboard_report.name_override:
            return self.dashboard_report.name_override
        else:
            return self.report.name

    # noinspection PyMethodMayBeStatic
    def queries_menu(self):
        return []

    def get_dashboard_class(self, report):
        return None


class CustomModal(ModelFormModal):
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = CustomReport

    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name',
                   'report_tags',
                   'notes']



from ajax_helpers.mixins import AjaxHelpers
from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic import TemplateView
from django_menus.menu import MenuMixin, MenuItemDisplay
from django_modals.helper import modal_button, modal_button_method
from django_modals.modals import Modal

from advanced_report_builder.duplicate import DuplicateReport
from advanced_report_builder.models import Report
from advanced_report_builder.utils import split_slug
from advanced_report_builder.views.bar_charts import BarChartView
from advanced_report_builder.views.datatables.datatables import TableView
from advanced_report_builder.views.funnel_charts import FunnelChartView
from advanced_report_builder.views.kanban import KanbanView
from advanced_report_builder.views.line_charts import LineChartView
from advanced_report_builder.views.pie_charts import PieChartView
from advanced_report_builder.views.single_values import SingleValueView


class ViewReportBase(AjaxHelpers, MenuMixin, TemplateView):
    model = Report
    views = {'tablereport': TableView,
             'singlevaluereport': SingleValueView,
             'barchartreport': BarChartView,
             'linechartreport': LineChartView,
             'piechartreport': PieChartView,
             'funnelchartreport': FunnelChartView,
             'kanbanreport': KanbanView,
             }
    enable_links = True

    custom_views = {}

    views_overrides = {}

    def __init__(self, *args, **kwargs):
        self.report = None
        super().__init__(*args, **kwargs)

    def redirect_url(self):
        """ used if the slug changes"""
        return None

    def dispatch(self, request, *args, **kwargs):
        slug = split_slug(self.kwargs['slug'])
        self.report = self.model.objects.filter(slug=slug['pk']).first()

        if self.report is None:
            self.report = self.model.objects.filter(slug_alias=slug['pk']).first()
            if self.report is None:
                return self.report_not_found()
            else:
                redirect_url = self.redirect_url()
                if redirect_url:
                    return redirect_url

        if not self.has_permission():
            return self.report_no_permission()

        return super().dispatch(request, *args, **kwargs)

    def has_permission(self):
        """You can over override this to check if the user has permission to view the report.
          If return false 'report_no_permission' will be called"""
        return True

    def report_no_permission(self):
        raise Http404

    def report_not_found(self):
        raise Http404

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        view = self.get_view(report=self.report)
        self.kwargs['report'] = self.report
        self.kwargs['enable_links'] = self.enable_links
        report_data = view.as_view()(self.request, *self.args, **self.kwargs).rendered_content
        context['report'] = report_data
        return context

    def get_view(self, report):
        if report.instance_type == 'customreport':
            view_name = report.customreport.view_name
            return self.custom_views.get(view_name)
        elif report.instance_type in self.views_overrides:
            return self.views_overrides.get(report.instance_type)
        return self.views.get(report.instance_type)

    def post(self, request, *args, **kwargs):
        view = self.get_view(report=self.report)
        self.kwargs['report'] = self.report
        self.kwargs['enable_links'] = self.enable_links
        return view.as_view()(self.request, *self.args, **self.kwargs)


class DuplicateReportModal(Modal):
    menu_display = MenuItemDisplay('Duplicate', font_awesome='fas fa-copy')

    def modal_content(self):
        return 'Are you sure you want to duplicate this report?'

    def get_modal_buttons(self):
        return [modal_button_method('Confirm', 'duplicate'), modal_button('Cancel', 'close', 'btn-secondary')]

    def button_duplicate(self, **_kwargs):
        report = get_object_or_404(Report, id=self.slug['pk'])
        duplicate_report = DuplicateReport()
        new_report = duplicate_report.duplicate(report=report)
        messages.add_message(self.request, messages.SUCCESS, f'Successfully duplicated {report.name}')

        url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
        if url_name:
            url = reverse(url_name, kwargs={'slug': new_report.slug})
            return self.command_response('redirect', url=url)
        else:
            return self.command_response('reload')
